from pdm_polylith_workspace.polylith.repo.get import get_authors, get_python_version
from pdm_polylith_workspace.polylith.repo.repo import bases_dir, components_dir, default_toml, development_dir, get_workspace_root, is_hatch, is_pdm, is_pep_621_ready, is_poetry, projects_dir, readme_file, workspace_file
__all__ = ['get_authors', 'get_python_version', 'bases_dir', 'components_dir', 'default_toml', 'development_dir', 'get_workspace_root', 'is_hatch', 'is_pdm', 'is_pep_621_ready', 'is_poetry', 'projects_dir', 'readme_file', 'workspace_file']


import ast
from pathlib import Path
from typing import List, Union


def create_namespace_path(top_ns: str, current: str) -> str:
    top_ns_module_path = top_ns.replace("/", ".")
    return f"{top_ns_module_path}.{current}"


def mutate_import(node: ast.Import, ns: str, top_ns: str) -> bool:
    did_mutate = False

    for alias in node.names:
        if alias.name == ns:
            alias.name = create_namespace_path(top_ns, alias.name)
            did_mutate = True

    return did_mutate


def mutate_import_from(node: ast.ImportFrom, ns: str, top_ns: str) -> bool:
    did_mutate = False

    if not node.module or node.level != 0:
        return did_mutate

    if node.module == ns or node.module.startswith(f"{ns}."):
        node.module = create_namespace_path(top_ns, node.module)
        did_mutate = True

    return did_mutate


def mutate_imports(node: ast.AST, ns: str, top_ns: str) -> bool:
    if isinstance(node, ast.Import):
        return mutate_import(node, ns, top_ns)

    if isinstance(node, ast.ImportFrom):
        return mutate_import_from(node, ns, top_ns)

    return False


def rewrite(source: Path, ns: str, top_ns: str) -> bool:
    file_path = source.as_posix()

    with open(file_path, "r", encoding="utf-8") as f:
        tree = ast.parse(f.read(), source.name)

    res = {mutate_imports(node, ns, top_ns) for node in ast.walk(tree)}

    if True in res:
        rewritten_source_code = ast.unparse(tree)  # type: ignore[attr-defined]

        with open(file_path, "w", encoding="utf-8", newline="") as f:
            f.write(rewritten_source_code)

        return True

    return False


def rewrite_module(module: Path, ns: str, top_ns: str) -> Union[str, None]:
    was_rewritten = rewrite(module, ns, top_ns)

    return f"{module.parent.name}/{module.name}" if was_rewritten else None


def rewrite_modules(path: Path, ns: str, top_ns: str) -> List[str]:
    """Rewrite modules in bricks with new top namespace

    returns a list of bricks that was rewritten
    """

    modules = path.glob("**/*.py")

    res = [rewrite_module(module, ns, top_ns) for module in modules]

    return [r for r in res if r]



from pdm_polylith_workspace.polylith.toml.core import get_custom_top_namespace_from_polylith_section, get_project_dependencies, get_project_package_includes, get_project_packages_from_polylith_section, read_toml_document
__all__ = ['get_custom_top_namespace_from_polylith_section', 'get_project_dependencies', 'get_project_package_includes', 'get_project_packages_from_polylith_section', 'read_toml_document']


from pdm_polylith_workspace.polylith.configuration.core import get_brick_structure_from_config, get_namespace_from_config, get_resources_structure_from_config, get_tag_pattern_from_config, get_tests_structure_from_config, get_theme_from_config, is_readme_generation_enabled, is_test_generation_enabled
__all__ = ['get_brick_structure_from_config', 'get_namespace_from_config', 'get_resources_structure_from_config', 'get_tag_pattern_from_config', 'get_tests_structure_from_config', 'get_theme_from_config', 'is_readme_generation_enabled', 'is_test_generation_enabled']


from functools import lru_cache
from pathlib import Path
import tomlkit
from pdm_polylith_workspace.polylith.repo.repo import default_toml, is_pep_621_ready

@lru_cache
def get_pyproject_data(path: Path) -> tomlkit.TOMLDocument:
    with open(str(path / default_toml), 'r', errors='ignore') as f:
        return tomlkit.loads(f.read())

def get_metadata_section(data: dict) -> dict:
    return data['project'] if is_pep_621_ready(data) else data['tool']['poetry']

def get_authors(path: Path) -> str:
    data = get_pyproject_data(path)
    section = get_metadata_section(data)
    return section.get('authors', []).as_string()

def get_python_version(path: Path) -> str:
    data: dict = get_pyproject_data(path)
    if is_pep_621_ready(data):
        return data['project'].get('requires-python', '')
    return data['tool']['poetry']['dependencies']['python']


import re
from pathlib import Path
from typing import List, Union
import tomlkit
from pdm_polylith_workspace.polylith import repo

def transform_to_package(namespace: str, include: str) -> dict:
    path, _separator, brick = str.partition(include, f'/{namespace}/')
    return {'include': f'{namespace}/{brick}', 'from': path}

def get_polylith_section(data) -> dict:
    return data.get('tool', {}).get('polylith', {})

def get_custom_top_namespace_from_polylith_section(data) -> Union[str, None]:
    poly_data = get_polylith_section(data)
    return poly_data.get('build', {}).get('top-namespace')

def get_project_packages_from_polylith_section(data) -> dict:
    poly_data = get_polylith_section(data)
    bricks = poly_data.get('bricks')
    return bricks if isinstance(bricks, dict) else {}

def get_hatch_project_packages(data) -> dict:
    polylith_section = get_project_packages_from_polylith_section(data)
    if polylith_section:
        return polylith_section
    hatch_data = data['tool']['hatch']
    build_data = hatch_data.get('build', {}) if isinstance(hatch_data, dict) else {}
    return build_data.get('force-include', {})

def get_project_package_includes(namespace: str, data) -> List[dict]:
    if repo.is_poetry(data):
        return data['tool']['poetry'].get('packages', [])
    includes = get_hatch_project_packages(data) if repo.is_hatch(data) else get_project_packages_from_polylith_section(data)
    return [transform_to_package(namespace, key) for key in includes.keys()]

def get_project_name(data) -> str:
    if repo.is_pep_621_ready(data):
        return data['project']['name']
    return data['tool']['poetry']['name']

def get_project_dependencies(data) -> dict:
    if repo.is_poetry(data):
        deps = data['tool']['poetry'].get('dependencies', [])
        items = set(deps.keys())
    else:
        deps = data['project'].get('dependencies', [])
        items = {re.split('[\\^~=!<>]', dep)[0] for dep in deps}
    return {'items': items, 'source': repo.default_toml}

def read_toml_document(path: Path) -> tomlkit.TOMLDocument:
    with path.open(encoding='utf-8', errors='ignore') as f:
        return tomlkit.loads(f.read())


import shutil
from pathlib import Path
from pdm_polylith_workspace.polylith import parsing

def get_work_dir(config: dict) -> Path:
    build_config = config.get('tool', {}).get('pdm', {}).get('build', {})
    work_dir = build_config.get('work-dir', '.polylith_tmp')
    return Path(work_dir)

def copy_bricks_as_is(bricks: dict, build_dir: Path) -> None:
    for source, brick in bricks.items():
        parsing.copy_brick(source, brick, build_dir)

def copy_and_rewrite_bricks(bricks: dict, top_ns: str, work_dir: Path, build_dir: Path) -> None:
    ns = parsing.parse_brick_namespace_from_path(bricks)
    for source, brick in bricks.items():
        path = parsing.copy_brick(source, brick, work_dir)
        rewritten_bricks = parsing.rewrite_modules(path, ns, top_ns)
        destination_dir = build_dir / top_ns
        parsing.copy_brick(path.as_posix(), brick, destination_dir)
        for item in rewritten_bricks:
            print(f'Updated {item} with new top namespace for local imports.')

def cleanup(work_dir: Path) -> None:
    if not work_dir.exists() or not work_dir.is_dir():
        return
    shutil.rmtree(work_dir.as_posix())


from functools import lru_cache
from pathlib import Path
from typing import Union
import tomlkit
from pdm_polylith_workspace.polylith import repo

@lru_cache
def _load_workspace_config(path: Path) -> tomlkit.TOMLDocument:
    fullpath = path / repo.workspace_file
    if not fullpath.exists():
        fullpath = path / repo.default_toml
    content = fullpath.read_text()
    return tomlkit.loads(content)

def get_namespace_from_config(path: Path) -> str:
    toml: dict = _load_workspace_config(path)
    return toml['tool']['polylith']['namespace']

def get_git_tag_pattern(toml: dict) -> str:
    """Fallback git tag pattern configuration"""
    return toml['tool']['polylith']['git_tag_pattern']

def get_tag_pattern_from_config(path: Path, key: Union[str, None]) -> str:
    toml: dict = _load_workspace_config(path)
    patterns = toml['tool']['polylith'].get('tag', {}).get('patterns')
    return patterns[key or 'stable'] if patterns else get_git_tag_pattern(toml)

def is_test_generation_enabled(path: Path) -> bool:
    toml: dict = _load_workspace_config(path)
    enabled = toml['tool']['polylith']['test']['enabled']
    return bool(enabled)

def is_readme_generation_enabled(path: Path) -> bool:
    toml: dict = _load_workspace_config(path)
    enabled = toml['tool']['polylith'].get('resources', {}).get('brick_docs_enabled')
    return bool(enabled)

def get_theme_from_config(path: Path) -> str:
    toml: dict = _load_workspace_config(path)
    return toml['tool']['polylith']['structure'].get('theme') or 'tdd'

def get_brick_structure_from_config(path: Path) -> str:
    theme = get_theme_from_config(path)
    if theme == 'loose':
        return '{brick}/{namespace}/{package}'
    return '{brick}/{package}/src/{namespace}/{package}'

def get_tests_structure_from_config(path: Path) -> str:
    theme = get_theme_from_config(path)
    if theme == 'loose':
        return 'test/{brick}/{namespace}/{package}'
    return '{brick}/{package}/test/{namespace}/{package}'

def get_resources_structure_from_config(path: Path) -> str:
    theme = get_theme_from_config(path)
    if theme == 'loose':
        return '{brick}/{namespace}/{package}'
    return '{brick}/{package}'


from pdm_polylith_workspace.polylith.parsing.core import copy_brick, parse_brick_namespace_from_path
from pdm_polylith_workspace.polylith.parsing.rewrite import rewrite_modules
__all__ = ['copy_brick', 'parse_brick_namespace_from_path', 'rewrite_modules']


from pathlib import Path
from typing import Set
from pdm_polylith_workspace.polylith import configuration, toml

def paths_from_config(ns: str, data: dict) -> Set[str]:
    packages = toml.get_project_package_includes(ns, data)
    return {p['from'] for p in packages}

def parse_paths(root: Path, theme: str, ns: str, data: dict) -> Set[str]:
    defaults = {'bases', 'components'}
    paths = defaults if theme == 'loose' else paths_from_config(ns, data)
    return {(root / p).as_posix() for p in paths}

def write_pth_file(build_dir: Path, paths: Set[str]) -> None:
    filepath = build_dir / 'polylith_workspace.pth'
    with open(filepath, 'w') as f:
        for p in paths:
            f.write(f'{p}\n')

def build_initialize(config_data: dict, build_dir: Path, root: Path) -> None:
    theme = configuration.get_theme_from_config(root)
    ns = configuration.get_namespace_from_config(root)
    paths = parse_paths(root, theme, ns, config_data)
    if not paths:
        return
    write_pth_file(build_dir, paths)


import shutil
from pathlib import Path


def copy_tree(source: str, destination: str) -> Path:
    ignore = shutil.ignore_patterns(
        "*.pyc",
        "__pycache__",
        ".venv",
        "__pypackages__",
        ".mypy_cache",
        ".pytest_cache",
        "node_modules",
        ".git",
    )

    res = shutil.copytree(source, destination, ignore=ignore, dirs_exist_ok=True)

    return Path(res)


def copy_brick(source: str, brick: str, destination_dir: Path) -> Path:
    destination = Path(destination_dir / brick).as_posix()

    return copy_tree(source, destination)


def parse_brick_namespace_from_path(bricks: dict) -> str:
    parts = {str.split(v, "/")[0] for v in bricks.values()}

    return next(part for part in parts)



from pathlib import Path
from pdm_polylith_workspace.polylith.pdm.hooks.workspace import build_initialize

def pdm_build_initialize(context):
    """Adding an additional pth file to the virtual environment

    Making the virtual environment aware of the Polylith Workspace.
    """
    context.ensure_build_dir()
    data = context.config.data
    build_dir = Path(context.build_dir)
    root = Path(context.config.root)
    build_initialize(data, build_dir, root)


from pathlib import Path
from pdm_polylith_workspace.polylith import toml
from pdm_polylith_workspace.polylith.pdm import core

def build_initialize(config_data: dict, build_dir: Path) -> None:
    bricks = toml.get_project_packages_from_polylith_section(config_data)
    top_ns = toml.get_custom_top_namespace_from_polylith_section(config_data)
    work_dir = core.get_work_dir(config_data)
    if not bricks:
        print('No bricks found.')
        return
    if not top_ns:
        core.copy_bricks_as_is(bricks, build_dir)
    else:
        core.copy_and_rewrite_bricks(bricks, top_ns, work_dir, build_dir)
        core.cleanup(work_dir)